const util = require('util')
const {copy, flat} = require('utils')

const initializedKey = Symbol('initialized')
const hooksKey = Symbol('hooks')
const declaresKey = Symbol('declares')
const subscribersKey = Symbol('subscribers')

const triggerHooksKey = Symbol('triggerHooks')
const addHooksKey = Symbol('addHooks')

class Pluggable {

  constructor () {
    if (this.constructor === Pluggable) {
      throw ReferenceError('Pluggable is an interface.')
    }
    if (!this.constructor[initializedKey]) {
      Pluggable.init(this.constructor)
    }
    this[hooksKey] = copy(this.constructor[hooksKey])
    this.constructor[subscribersKey].push(this)
    for (let dec of this.constructor[declaresKey]) {
      this[dec.path] = typeof dec.value === 'function' ? dec.value() : copy(dec.value)
    }
  }

  /**
   * @description Add a pre plugin.
   * @param {string|Array.<string>} name          - Hook's name.
   * @param {...(function|Array.<function>)} func - Plugin's functions.
   */
  pre (name, ...func) {
    this[addHooksKey]('pre', name, ...func)
  }

  /**
   * @description Add a post plugin.
   * @param {string|Array.<string>} name          - Hook's name.
   * @param {...(function|Array.<function>)} func - Plugin's functions.
   */
  post (name, ...func) {
    this[addHooksKey]('post', name, ...func)
  }

  /**
   * @param {"pre"|"post"}period
   * @param {string}name
   * @param {...*}argus
   * @private
   */
  async [triggerHooksKey] (period, name, ...argus) {
    const plugins = this[hooksKey][period][name]
    if (plugins) {
      for (let plugin of plugins) {
        await plugin.apply(this, argus)
      }
    }
  }

  /**
   *
   * @param {"pre"|"post"}period
   * @param {string|Array.<string>}name
   * @param {...(function|Array.<function>)}func
   * @private
   */
  [addHooksKey] (period, name, ...func) {
    const hooks = this[hooksKey][period]
    if (typeof name === 'string') {
      hooks[name].push(...flat(func))
    }
    else if (Array.isArray(name)) {
      for (let n of name) {
        hooks[n].push(...flat(func))
      }
    } else {
      throw ReferenceError('Hook\'s name is string or array of string')
    }
  }

  /**
   * @description Add a pre plugin to all instance of the class.
   * @see pre
   */
  static pre () {
    Pluggable.prototype.pre.apply(this, arguments)
  }

  /**
   * @description Add a post plugin to all instance of the class.
   * @see post
   */
  static post () {
    Pluggable.prototype.post.apply(this, arguments)
  }

  /**
   * @private
   */
  static [addHooksKey] () {
    Pluggable.prototype[addHooksKey].apply(this, arguments)
    for (let subscriber of this[subscribersKey]) {
      Pluggable.prototype[addHooksKey].apply(subscriber, arguments)
    }
  }

  /**
   *
   * @param Class
   */
  static init (Class) {
    Class[initializedKey] = true
    Class[hooksKey] = {
      pre: {},
      post: {}
    }
    Class[declaresKey] = []
    Class[subscribersKey] = []
  }

  /**
   * @description Declare a property for all instance for this class.
   * @param {string}path    - Name of the property.
   * @param {*}value        - Initial value (or value generator) for the property.
   */
  static declare (path, value) {
    this[declaresKey].push({path, value})
    for (let subscriber of this[subscribersKey]) {
      subscriber[path] = typeof value === 'function' ? value() : copy(value)
    }
  }

  /**
   * @description Wrap a pluggable method.
   * @see readme.md
   * @param {string}funcName
   * @param {string}hookName
   * @param {string?}periodsSet
   */
  static wrapHook (funcName, hookName, periodsSet) {
    if (this === Pluggable) {
      throw ReferenceError('Pluggable is an interface.')
    }
    if (!this[initializedKey]) {
      Pluggable.init(this)
    }
    if (!periodsSet) {
      periodsSet = hookName
      hookName = funcName
    }
    const periods = periodsSet.split(' ')
      .map(e => e.trim())
      .filter(e => /pre|post/.test(e))
    if (periods.length !== 0) {
      const hooks = this[hooksKey]
      periods.forEach(period => {
        hooks[period][hookName] = []
      })
      const originalMethod = this.prototype[funcName]
      this.prototype[funcName] = async function (...args) {
        let result = await this[triggerHooksKey]('pre', hookName, ...args)
          .then(async () => typeof originalMethod === 'function' ? await originalMethod.apply(this, args) : args[0])
        await this[triggerHooksKey]('post', hookName, result)
        return result
      }
    } else {
      throw SyntaxError(`wrapped methods needs periods.`)
    }
  }
}

/**
 * @type {Pluggable}
 */
module.exports = Pluggable
